#define SDRAM 1
#include <def21161.h>

/* -----------------------------------------------------------------------

        161_link.asm

        link-port based boot loader of the ADSP-21161 processor.

        Copyright (c) 2001 Analog Devices Inc.  All rights reserved.
                6/06/01

        Modifications:
                28.FEB.2002 - M.Walsh
                -Added conditional SDRAM initialization
                -Removed redundant initializations of several DAG registers
                -Added comments including tag information
                -Added call to "user_init" subroutine
                -Relocated "read_link_word" above P0I to make room for user init
                -Optimized (for size) all four ext_code_init code sections
                        (code that processes tags 0x13, 0x14, 0x15, and 0x16)
                -Collected all the code for all four external code init sections
                        into into this single source file 161_link.asm, eliminating
                        the previous need for four kernel.DXE's
                -Added a single ext_code_zero initialization routine that is used
                        to zero-out all four sections with ZERO_PM*_EXT tags.
                        External count is now derived from actual tag 0x17, 0X18,
                        0x19, and 0X1A.  IF TAGS EVER CHANGE, SO MUST THIS INIT-CODE .
                -Now handle PM16, DM16, PM32, and DM32 with single init routine
                -Now handle PM40, DM40, and PM48 with single init routine

                01.MAR.2002 - M.Walsh
                -Added conditional debugging capability.  By #defining PAUSE, and then
                        rebuilding this kernel, the instruction "jump(pc,0);" is included.
                        This will cause the kernel to be booted and begin executing, but
                        it will not do so completely.  If "Initial Reset on Startup"
                        is NOT enabled for your emulator session, you can attach
                        and debug the kernel.

                03.JUN.2002 - M.Walsh
                -IMASK now cleared after final init.
                -Removed EXT_PM48_INIT code because the linkports are multiplexed with the
                 lower-16 data pins which are required for x48 memory. They cannot be used
                 simultaneously.
                -Moved last six initialization TAG's above LP0I to conserve space.

                20.DEC.2002 - M.Walsh
                -Changed #define SDRAM 1 to #define SDRAM 0

                16.JAN.2003 - M.Walsh/A.Caldwell
                -Added instruction in EXT_ZERO_CODE to trap tags >= 0x1A.  This tag
                    (x48 ext. zero-init) is incompatible w/ linkport booting since
                    the linkports are MUX'd with the D0-D15.  This is the highest tag,
                    so all greater tags are illegal.



IMPORTANT NOTES:
1) This kernel uses Link Port Buffer Zero to import data.  The four instructions
that handle the associated interrupt MUST be located at 0x38 in the interrupt
vector table.

2) Unused interrupt vectors are used for additional loader kernel space
due to the new TAGs added to the ADSP-21161.  This helps save space for
additional customization of the kernel by the user.  Out of the reset vector,
IMASK and LIRPTL are cleared to ensure that these interrupts will not be serviced.

3) Once booting completes, Linkport buffer zero is left enabled!

4) 48-bit wide external memory is not supported for linkport booting because
the linkport pins are multiplexed with the lower 16 data pins.
---------------------------------------------------------------------------*/

.section /pm seg_ldr;
                nop;nop;nop;nop;

reset:          IDLE;                               // Implicit IDLE instruction at 0x40004

// ------------------ DEBUG BOOTING USING ICE? ---------------------------------------------------------
// Include this "trap" instruction by using the -D switch on the commandline when building 161_link.dxe.
// (for example: 	easm21k.exe -21161 -o 161_prom.doj 161_prom.asm -DPAUSE)
//
// Attach emulator probe and hit run.  While running, reset the board to trigger LINKPORT booting.
// Now, when you hit halt with the emulator, the kernel will be loaded, and you'll be at the following
// instruction.  Change this instruction to a "NOP" in the disassembly window, and you'll be able to
// single-step through the kernel.
#ifdef PAUSE
    jump (pc,0);    // "pause" execution of boot-kernel to allow ICE to break in after booting begins.
#else
    nop;
#endif
// (Note: to make debugging the kernel more simple, you can load the kernel's LABEL's using the
// following TCL command in the IDDE console:
//
//                  dspload "C:\\<path>\\161_prom.DXE" -symbols
//
// ------------------------------------------------------------------------------------------------------


                JUMP loader_kernel_init (DB);       // 1st instruction executed after reset
                IMASK = 0;                          // mask all interrupts, since we use unused vectors for kernel instructions to save space
                LIRPTL = 0;

loader_kernel_init:
                CALL user_init;                     // before executing the kernel, execute user's system initialization code

                // Zero out L-registers so I-registers can be used without wrap
                L0=0; L4=0;     //DAG1
                L8=0; L12=0;    //DAG2

                //Init M-registers
                M5=0; M6=1;     //DAG1
                M13=0; M14=1;   //DAG2

                R8 = 2;                 // Used to decrement tag by two during tag evaluation
        		R12=DM(SYSCON);	        // Read current SYSCON setting
        		R11=R12;				// For setting IMDW bit later

                I15 = 0x40004;
                R9 = L0EN|L0EXT|LAB1;		// Lbuf0 enabled, DMA disabled, receive, 48bits
                dm(LCTL) = R9; 			// Lbuf0 enabled, DMA disabled, receive, 48bits


// ==========================  READ_BOOT_INFO  =======================================
// Places TAG in R0, Internal Count in R2, and Destination Address in R3
// -----------------------------------------------------------------------------------
read_boot_info: CALL read_link_word;    // read first tag and count
                R0=PASS R2;             // r0 holds tag
                CALL read_link_word;
                jump Start_Load;
// ========================= END READ_BOOT_INFO =======================================



// ==========================  read_link_word  =======================================
// read_link_word is a callable subroutine that DMA's in one 48-bit word from PROM.
// It places the MS 32-bits in R3 and the LS 16 (right justified) in R2.
// For example, given the 48-bit word 0x112233445566,
//        R2 holds 0x00005566
//        R3 holds 0x11223344
// -----------------------------------------------------------------------------------
read_link_word: ustat1=dm(LCTL);			// Wait for LBUF0 to have data
				bit tst ustat1 0x00800000;
				if not TF jump read_link_word;

				PX=PM(LBUF0);				// Read word from LBUF0
				R2=PX1;						// Copy PX values into DREGS
				RTS (DB);
    				R2=LSHIFT R2 by -16;		// Downshift 16 bits to adjust for the extension of PX to 64-bits
	    			R2=PASS R2, R3=PX2;			// Test the length
// ======================== END read_link_word =======================================



//=====================================================================================
// Last six TAG's are handled here before the LP0I interrupt vector to conserve space:
//=====================================================================================
//--------------- 21 0X0015 INIT PM32 EXT ---------------
test_pm32_initext:      R0=R0-R8;           // R8 = 2
                        IF EQ JUMP (PC,0);          // TAG == 0x16 which is not supported in linkport booting. See Note #4 above.
                        IF GT JUMP ext_zero_code;   // TAG >= 0x17
pm32_init_ext_code:     I0=R3;                      //else TAG==0x15 use this init
                        LCNTR=R2, DO pm32_init_external_loop UNTIL LCE;
                        CALL read_link_word;
                        DM(I0,M6)=R3;
                        DM(I0,M6)=R2;
pm32_init_external_loop:nop;            // call can't be in last 3 inst. of loop
                        JUMP read_boot_info;

//---------------------------- EXT_ZERO_CODE ---------------------------------
// This portion of code is used to initialize external memory holding NOP's
// (NOP opcode is 0x000000000000)
// r2 holds number of external #NOP's to be initialized
// r0 holds (TAG - 22) and is used to calculate the number of addresses per NOP.
//
//      If  pm8, r0 = 1, multiply r2 by 8      (tag: 23 0X0017 ZERO PM8 EXT)
//      If pm16, r0 = 2, multiply r2 by 4      (tag: 24 0X0018 ZERO PM16 EXT)
//      If pm32, r0 = 3, multiply r2 by 2      (tag: 25 0X0019 ZERO PM32 EXT)
//      If pm48, r0 = 4,    do nothing since 48-bit external memory is not supported
//                          under linkport booting (tag: 26 0X001A ZERO PM48 EXT)
//-----------------------------------------------------------------------------
ext_zero_code: 	r5=4;
				r0=r5-r0, I0=R3;	    // external count derived from tag (in r0), I0=ext address
				if GE jump (pc,0);      // trap if tag >= 0x1A (x48 ext. mem. incompatible w/ linkports)
				r2=lshift r2 by r0;	    // r2 = number addresses per NOP (shift by n == mult. by 2^n)
                LCNTR=R2, DO ext_zero_code.end UNTIL LCE;
ext_zero_code.end:      DM(I0,M6)=M13;  //m13 = 0
                JUMP read_boot_info;    //fetch header for next block
//=====================================================================================

nop;nop;nop;


// ============================== LPB0_ISR  =================================================
// DMA Channel 8 (LPB0) Complete. This code MUST resolve to 0x38 in the interrupt vector table.
// If any code is added or removed prior to this, the kernel will not function properly. Pad
// with NOP's as done above if needed.  (consult "kernel.map" for linking results)
lp0i_at_0x38:   nop;
                RTI (DB);
                bit clr lirptl 0x10000;     //mask link port interrupts
                nop;
// ============================= END LP0_ISR  =========================================




//====================================================================================
// ============================= FINAL_INIT ==========================================
// This code is the last piece run in the kernel.  It uses a DMA to initialize the
// 256 instructions that reside in the interrupt vector table.
// -----------------------------------------------------------------------------------
final_init:	R11=BSET R11 BY 9;      // Set IMDW to 1 for inst write
			DM(SYSCON)=R11;         // Setup for instruction write
            R9=0xb16b0000;          // Load instruction PM(0,I8)=PX;
			I4=0x040004;            // location of this instruction
            I8=0x040038;            // Point to DMA8 vector to patch

            //----------Initialize final DMA-------------
			R1=0x040000;            // Point to destination
			DM(IILB0)=R1;           // Setup DMA to load over ldr
			R2=0x100;               // Load length of last init
			DM(CLB0)=R2;			// Load internal count
			DM(CPLB0)=M5;			// Set CP to 0
    		DM(IMLB0)=M14;          // Set to increment internal ptr
            //-------------------------------------------

			R9 = pass R9, R11=R12;	// Clear AZ, hold initial SYSCON
			DO 0x40004 UNTIL EQ;      // Setup dummy loop
			PCSTK=0x40004;			// replace w/ new top-of-loop value
            FLUSH CACHE;			// Flush cache to prevent any false cache-hits once the kernel is overwritten

            r8=L0EN|L0DEN|L0EXT|LAB1;
			DM(LCTL)=R8;			// Begin final DMA transfer (256-instructions)

			bit set mode1 0x1800;   // allow global/nested interrupts
			bit clr lirptl 0x1;		// clear any pending LP0 interrupts
			bit set lirptl 0x10000;	// enable LP0 interrupt
			bit set imask 0x4000;	// enable LPISUM interrupt

			JUMP 0x40004 (DB);      // Jump to start
    			IDLE;			    // Wait for final DMA to complete before "landing" at 0x40004.
                IMASK=0;            // Mask all interrupts once this final DMA completes

// When this final DMA completes, IDLE in the delayed branch completes, IMASK is cleared,
// and then the DSP executes the instruction at 0x40038 - the vector for LP0I. The loader
// has automatically placed an RTI at that address, which, when executed, sends the processor
// to 0x40004 (because it's been pushed onto the PCSTACK). The loader has automatically placed
// one more instruction, and that is at 0x40004:
//
//          R0=R0-R0, DM(I4,M5)=R9, PM(I12,M13)=R11;   //opcode: x39732D802000
//
// Executing this instruction performs the three things:
// 1) It sets the AN flag ("until EQ" now evaluates as true)
// 2) It restores the power-up value of SYSCON (held in R11).
// 3) It overwrites itself with the instruction PM(0,I8)=PX; (held in R9)
//
// Because the termination condition (EQ) is achieved within the last three instructions
// of the loop, the DSP jumps to the top-of-loop address one last time.  We manually changed this
// top-of-loop address (PCSTK) to x40004, so, to conclude the kernel, the DSP executes the instruction
// at 0x40004 again.  The new instruction there resets the LP0I interrupt vector (x40038) to it's
// user-intended final value (since the loader inserted an RTI there). At this point, the kernel
// is finished, and execution continues at x40005 as if nothing happened!
//
//===================================== END FINAL_INIT =================================
//======================================================================================



// =============================  USER_INIT   ========================================
// If any registers need initialization before the kernel begins importing the rest of
// the application, they should be done so here.  Typically a user would need to
// initialize SDRAM settings, but they may also modify WAIT, SYSCON, SPICTL, or LCTL to
// increase the speed of the boot process.
//
// This subroutine may include NO MORE THAN 88 instructions (in addition to the RTS).
// -----------------------------------------------------------------------------------
user_init:

#if SDRAM
                r0=0x01C00000;      // clear MSx waitstate and mode
                dm(WAIT)=r0;
                r0=0x1000;          // refresh rate
                dm(SDRDIV)=r0;
                r0=0x02014231;      // SDRAM running @ Core-clock
                dm(SDCTL)=r0;
#endif
                RTS;                // return from user_init to loader_kernel_init
// ============================  END  USER_INIT  =====================================



// =============================  START LOAD   ========================================
// The following code sections analyze the TAG for each data section and initialize
// accordingly.  READ_BOOT_INFO is always called immediately before START_LOAD, and that
// routine places the TAG in R0, the Word Count in R2, and the destination Address in R3.
//
// Notes:
// - Several tags are handled with the same initialization code to save space.
// - These instructions MUST execute sequentially because the tag value
//      is decremented after each false test.
// - The tag value in decimal and hex along with the tag name is commented above each
//      section of initialization code
// -----------------------------------------------------------------------------------
Start_Load:

//--------------- 0 0x0000 FINAL INIT ---------------
                R0=PASS R0;
                IF EQ JUMP (PC, final_init);            // if tag = zero, execute final init
// ---------------------------------------------------


//--------------- 1 0x0001 ZERO DM16 ---------------
//--------------- 2 0x0002 ZERO DM32 ---------------
test_dm_zero:   R0=R0-R8;   //
                IF GT JUMP (PC, test_dm40_zero);        // jump unless tag == 2 or tag==1
dm_zero:        I0=R3;                                  // I0 holds SW or NW address accordingly
                LCNTR=R2, DO dm_zero_loop UNTIL LCE;    // R2 holds number of SWs or NWs
dm_zero_loop:   DM(I0,M6)=M13;
                JUMP read_boot_info;
// ---------------------------------------------------

//--------------- 3 0x0003 ZERO DM40 ---------------
test_dm40_zero: R0=R0-1;
                IF NE JUMP (PC, test_dm_init);          // jump unless tag == 3
dm40_zero:      PX=0;
                I0=R3;
                LCNTR=R2, DO dm40_zero.end UNTIL LCE;
dm40_zero.end:  DM(I0,M6)=PX;
                JUMP read_boot_info;    //fetch header for next block
// -------------------------------------------------------


//--------------- 4 0x0004 INIT DM16 ---------------
//--------------- 5 0x0005 INIT DM32 ---------------
test_dm_init:   R0=R0-r8;   // r8 = 2
                IF GT JUMP (PC, test_dm40_init);        // jump unless tag == 4 or tag == 5
dm_init:        I0=R3;
                LCNTR=R2, DO dm_init.end UNTIL LCE;
                CALL read_link_word(DB);
                NOP; NOP;               //call cannot be in last 3 inst. of a loop
dm_init.end:    DM(I0,M6)=R3;
                JUMP read_boot_info;    //fetch header for next block
// -------------------------------------------------------


//--------------- 6 0x0006 INIT DM40 ---------------
test_dm40_init: R0=R0-1;
                IF NE JUMP (PC, test_pm_zero);            // jump unless tag == 6
dm40_init:      I0=R3;
                LCNTR=R2, DO dm40_init.end UNTIL LCE;
                CALL read_link_word (DB);NOP;NOP;   //call cannot be in last 3 inst. of a loop
dm40_init.end:  DM(I0,M6)=PX;
                JUMP read_boot_info;    //fetch header for next block
// -------------------------------------------------------


// PM and DM zero-filled sections are handled by the same code:
//--------------- 7 0x0007 ZERO PM16 ---------------
//--------------- 8 0x0008 ZERO PM32 ---------------
//--------------- 9 0x0009 ZERO PM40 ---------------
//--------------- 10 0x000A ZERO PM48 ---------------
test_pm_zero:   R0=R0-R8;   //r8==2
                IF LE JUMP (PC, dm_zero);   // jump if tag==7 or tag==8
                R0=R0-R8;   //r8==2
                IF LE JUMP (PC, dm40_zero); // jump if tag==9 or tag==10

// PM and DM initializations are handled by the same code:
//--------------- 11 0x000B INIT PM16 ---------------
//--------------- 12 0x000C INIT PM32 ---------------
//--------------- 13 0x000D INIT PM40 ---------------
//--------------- 14 0x000E INIT PM48 ---------------
test_pm_init:   R0=R0-R8;  //r8==2
                IF LE JUMP (PC, dm_init);   //jump if tag==11 or 12
                R0=R0-R8;  //r8==2
                IF LE JUMP (PC, dm40_init); //jump if tag==13 or 14
//------------------------------------------------------------


//--------------- 15 0x000F ZERO DM64 ---------------
test_dm64_zero: R0=R0-1;
                IF NE JUMP (PC, test_dm64_init);
dm64_zero:      R2=R2+R2;               //double count to initialize in NW memory
                R3 = LSHIFT R3 by 1;    //convert LW address to NW address
                jump dm_zero;           //re-use code to conserve space
//------------------------------------------------------------


//--------------- 16 0x0010 INIT DM64 ---------------
test_dm64_init: R0=R0-1;
                IF NE JUMP (PC, test_pm64_zero);
dm64_init:      I0=R3;
                LCNTR=R2, DO dm64_init.end UNTIL LCE;
                CALL read_link_word;    // read 32 lsbs of 64-bit word
                R10=PX2;                // save 32 lsbs of 64-bit word
                CALL read_link_word;    // read 32 msbs of 64-bit word
                PX1=R10;                // restore 32 lsbs to PX1
                nop;                    // Call cannot be in last 3 instructions of end
dm64_init.end:  DM(I0,M6)=PX (lw);      // write all 64-bits to longword address location
                JUMP read_boot_info;    //fetch header for next block
//------------------------------------------------------------

//--------------- 17 0x0011 ZERO PM64 ---------------
//--------------- 18 0x0012 INIT PM64 ---------------
test_pm64_zero: R0=R0-R8;  //r8==2
                IF LT JUMP (PC, dm64_zero); // jump if tag==17
                IF EQ JUMP (PC, dm64_init); // jump if tag==18
                JUMP (PC,test_pm8_initext);
//------------------------------------------------------------

//-----------------------------------------------------------------------------------------
// "INIT PMx EXT" loads "packed" instructions into external memory.  External memory
// starts at address 0x200000 on the 21161. The format for this packing can be found
// in the 21161 Hardware Reference (at the end of Chapter 5).  Zero-init code sections
// (consecutive NOP's) are initialized seperatly in the EXT_ZERO_CODE subroutine.
// NOTE: Instructions may not be placed into x48-bit external memory when using linkports.
//-----------------------------------------------------------------------------------------

//--------------- 19 0x0013 INIT PM8 EXT ---------------
test_pm8_initext:       R0=R0-1;
                        IF NE JUMP test_pm16_initext;
pm8_init_ext_code:      I0=R3;
                        LCNTR=R2, DO pm8_init_external.end UNTIL LCE;
                        CALL read_link_word;    //r3= bits 47-16, r2= bits 15-0
                        //-------- write out 4 MSB's --------
                        lcntr=4, do (pc,2) until LCE;
                        r3=rot r3 by 8;
                        dm(i0,m6)= r3;
                        //-------- write out 2 LSB's --------
                        r4= LSHIFT R2 by -8;
                        dm(i0,m6) = r4; // write out 8-15
pm8_init_external.end:  dm(i0,3) = r2;  // write out 7-0, followed by 2 uninitialized locations
                        JUMP read_boot_info;    //fetch header for next block

//--------------- 20 0X0014 INIT PM16 EXT ---------------
test_pm16_initext:      R0=R0-1;
                        IF NE JUMP test_pm32_initext; // test_pm32_initext is located above
pm16_init_ext_code:     I0=R3;
                        LCNTR=R2, DO pm16_init_external_loop UNTIL LCE;
                        CALL read_link_word;
                        R5 = LSHIFT R3 BY -16;  //r3= bits 47-16, r2= bits 15-0
                        DM(I0,M6)=R5;
                        DM(I0,M6)=R3;   // "unpack" instruction opcode bits 31:16
pm16_init_external_loop:DM(I0,2)=R2;    // "unpack" instruction opcode bits 15:0, skip one location
                        JUMP read_boot_info;

//---------------------------------------------------------------------------
// TAGs 0x15 through 0x1A are handled in the unused interrupt vectors
// "above" the LPOI interrupt vector.
//---------------------------------------------------------------------------

//===========================================================================
//=============================================================================