/********************************************************************************************/
/*			ADSP-21161 Link port boot example												*/
/*			Code accompaniment to EE-199 Link Port Booting the ADSP-21161					*/
/*			Written by Andrew Caldwell														*/
/*			European DSP Applications														*/
/*			Analog Devices Inc																*/
/*																							*/
/*	Description:																			*/
/*		This program requires two ADSP-21161 EZ-KIT Lite development boards and a link port */
/*		cable to attach the two boards. It may also be used in a master DSP in an 			*/
/*		ADSP-21161 MP cluster.																*/
/*																							*/
/*		The file to be booted is a simple blink example that flashes LED 2,3,4,5,6,and 7	*/
/*		on the ADSP-21161N EZ-KIT Lite Evaluation Platform. These LEDs are connected to 	*/
/*		flags 9, 8, 7, 6, 5 and 4 respectively.												*/
/*																							*/
/*		The defined value "size" refers to the number of 16 bit words contained in the 		*/
/*		loader file. This is all that requires to be altered to boot with various loader	*/
/*		files. 																				*/
/*																							*/
/*		The loader file to be included can be built using either the ASCII or the Include 	*/
/*		file format, the example works with both.											*/
/*																							*/
/********************************************************************************************/ 
#include    "def21161.h"
#include	"defines.h"

.global __main;
/* Each 16-bit word of the loader file is stored in the lower 16 bits of 32-bit data memory */
.section/dm	seg_dmda;
.var _ldr_source[size] = "Blink_Example_SDRAM.ldr";	

.section/pm	seg_pmco;
__main:		
/* Clear link control registers*/
	r1 = 0x0;					
	dm(LCTL) = r1;
	
/* The 16 bit words from the _ldr_source buffer must be packed into a 48-bit instruction   */
/* before being transmitted. The loader file is of the format that the first 16-bit word  */
/* is the least significant 16 bit words of the 48-bit instruction, the next 16-bit word  */
/* is the middle 16-bits of the 48-bit instruction and the third 16-bit word is the 16	  */
/* most significant bits of the 48-bit instruction. This is the same for ASCII or Include */
/* loader file formats.																	  */


	B0 = _ldr_source; M0 = 1; L0 = 0;	// Declare base, modify and length values of buffer
	r0 = LAB1 | L0EN | L0TRAN |L0EXT;
	dm(LCTL) = r0;	

/* This routine calls the tx_instructions procedure passing the kernel_count of 256 to the routine */		
__transmit_kernel:
	call __tx_instructions (db);
	r4 = kernel_count;
	nop;
__transmit_kernel.end:


/* This routine calls the tx_instructions procedure passing the count value to the routine */
__transmit_code:
	call __tx_instructions (db);
	r4 = count;
	nop;
__transmit_code.end:
/* End of program*/
__stay: jump __stay;
__main.end:		

/* The tx_instructions routine is set up to run either kernel_count or count times to transmit  */
/* either the first 256 words or the remainder of the code.									 	*/
/* The loop consists of three parts. The first is test_lbuf. This reads the value of the link  	*/
/* port common control register to check if the link port buffer is full, if the buffer is full	*/
/* then it continually monitors this register until the buffer is not full. If the buffer is  	*/
/* not full then three 16-bit words are read and packed into the px register using px1 and px2 	*/
/* to create the 48-bit instruction to be transmitted.										   	*/
__tx_instructions:		
	lcntr = r4, do __tx_word until lce;	// set loop counter to 256 to transmit kernel

	__test_lbuf:
		r3 = dm(LCTL);					// read LCTL register
		r3 = ASHIFT r3 by LCTL_SHIFT;	// Shift so correct LBUF status can be compared to 0x3 (full)
		r2 = 0x3;					
		comp(r2,r3);
		if eq jump __test_lbuf;			// Check LBUF status, if full then test_lbuf, if not then create instruction
	__pack_word:
		r0 = dm(i0, m0);				// Read first 16-bit word
		r0 = LSHIFT r0 by 16;			// Shift 16 bits to MSBs of r0 to be loaded into px1
		px1 = r0;						// LSBs of px1 are dropped when reading px
		r0 = dm(i0, m0);				// Read middle 16 bits of instruction
		r1 = dm(i0, m0);				// Read upper 16 bits of instruction
		r0 = r0 OR LSHIFT r1 by 16;		// Create upper 32 bits of instruction
		px2 = r0;						// Store in px2
	__tx_word:
		pm(LBUF0) = px;				// Transmit contents of px. 48 bits are transmitted and not 64 (16 LSBs dropped)
	rts;
__tx_instructions.end:

		




